#!/bin/env python
# -*- coding: utf-8 -*-

import sys
import time

from king_snake.player import Player
from king_snake.chessboard import Chessboard
from king_snake.errors import ChessError

"""The GUI for playing chess using KingSnake!"""


class ChessGame(object):

    """
    A chess game manager

    What it should be able to do:
    Read a game from the command line
    Provide a user interface for moving pieces
    Prompt user to move pieces around
    """

    def __init__(self):
        player1, player2 = Player(), Player()
        self.white, self.black = player1, player2
        self.chessboard = Chessboard()
        self.chessboard.add_players(player1, player2)
        self.greet()

    def greet(self):
        """Greet the players and begin game."""
        print("Welcome to KingSnake.\n"
              "Enter moves directly or press Enter to enter the menu.")
        self.get_move()

    def get_move(self):
        """Show board and prompt for current move, then execute it."""
        valid_move = False
        while not valid_move:
            self.show()
            move = raw_input("Please enter your move (e.g. E2 E4) or enter to "
                             "access the menu: ")
            if not move:
                self.menu()
            else:
                try:
                    start_position, end_position = move.split()
                    self.chessboard.current_player.move(start_position.upper(),
                                                        end_position.upper())
                    valid_move = True
                except ValueError:
                    print("Please enter your move in the following form:\n"
                          "{start position} {end position}\n"
                          "Positions are notaded using their letter followed"
                          "by their number.\n"
                          "Example valid move request to move from A1 to A2: "
                          "'A1 A2'")
                except KeyError:
                    print("Only valid fields are allowed.")
                except ChessError as error:
                    print(error)
        self.get_move()

    def menu(self):
        """Allow user to do something other than move pieces"""

        def quit_game():
            """Quit game."""
            sys.exit()

        def resign():
            """Resign."""
            print("{} resigns. {} is the winner!".format(
                 self.chessboard.current_player.color.capitalize(),
                 self.chessboard.current_player.opponent.color.capitalize()))
            time.sleep(10)
            quit_game()

        def restart():
            """Start new game."""
            self.__init__()

        def save_game():
            """Save game to file."""
            raise NotImplementedError

        def load_game():
            """Load game from file."""
            raise NotImplementedError

        def undo_turn():
            """Undo turn."""
            raise NotImplementedError

        def return_to_game():
            """Resume play."""
            self.get_move()
        menu_choices = []
        for function in (quit_game, resign, restart, save_game, load_game,
                         undo_turn, return_to_game):
            menu_choices.append((function, function.__doc__))
        valid_choice = False
        while not valid_choice:
            for number, choice in enumerate(menu_choices):
                print("{}. {}".format(number + 1, choice[1]))
            try:
                decision = int(raw_input("What would you like to do?: ")) - 1
                try:
                    valid_choice = True
                    menu_choices[decision][0]()
                except IndexError:
                    valid_choice = False
                    print("Please enter a valid menu number.")
            except (ValueError, IndexError):
                print("Please enter a valid menu number.")

    def show(self):
        """Show chessboard and print current player."""
        print(self.chessboard)
        print("It's {}'s turn.\n".format(self.chessboard.current_player.color))

game = ChessGame()



"""A chess player."""

from king_snake.errors import (FieldMustBeCastledError,
                               FieldOccupiedError,
                               IllegalMoveError,
                               PawnMustCaptureError,
                               TurnError)
from king_snake.figures import Pawn, Rook, Knight, Bishop, Queen, King


class Player(object):

    """A chess player."""

    def __repr__(self):
        return "Player()"

    def __str__(self):
        if self.chessboard:
            if self == self.chessboard.players["white"]:
                color = "White"
            else:
                color = "Black"
            return_string = ("{color} Player on "
                             "{chessboard}\n"
                             "Figures: "
                             "{figures}".format(color=color,
                                                chessboard=self.chessboard,
                                                figures=self.figures))
        else:
            return_string = self.__repr__()
        return return_string

    @staticmethod
    def rollback(moved_piece, captured_piece=None):
        """
        Reverse a turn.

        @param moved_piece - Tuple of piece moved during the move with its old
                             position
        @param captured_piece - Tuple for captured piece, formated as above
        """
        rollback_pieces = [moved_piece]
        if captured_piece:
            rollback_pieces.append(captured_piece)

        for piece in rollback_pieces:
            piece["old_position"].receive_figure(piece["figure"])
            piece["figure"].position = piece["old_position"]

    def __init__(self):
        self.chessboard = None
        self.figures = None
        self.king = None
        self.color = None

    @property
    def opponent(self):
        """Return other player in chess game"""
        if self == self.chessboard.players["white"]:
            return self.chessboard.players["black"]
        else:
            return self.chessboard.players["white"]

    def set_up_board(self, chessboard):
        """Set up pieces on given chessboard and find other player."""
        self.chessboard = chessboard
        self.figures = list(Pawn(self) for pawns in range(8))
        for doubled_piece in (Rook, Knight, Bishop) * 2:
            self.figures.append(doubled_piece(self))
        self.figures.append(Queen(self))
        self.king = King(self)
        self.figures.append(self.king)

    def move(self, start_field, goal_field):
        """
        Move a piece to a new field.

        First verify if self is the chessboard's current player. Then check if
        a moveable figure is located at the start field. If the piece can be
        moved, move to the goal field, capturing a figure at the goal field if
        necessary. Finally, check if the move would put the own king in check.
        If yes, roll back the move. Otherwise, record the current turn on all
        moved pieces and end the turn.

        @param start_field - String used to look up field object (e.g. "E2")
        @param goal_field - Like start_field
        """
        if self != self.chessboard.current_player:
            raise TurnError("Move attempted out of turn.")

        start_field = self.chessboard.fields[start_field]
        goal_field = self.chessboard.fields[goal_field]

        figure = start_field.figure
        if not figure in self.figures:
            raise IllegalMoveError("Player does not own a piece at given "
                                   "position.")

        try:
            figure.move(goal_field)
            captured_piece = None
        except (FieldOccupiedError, PawnMustCaptureError):
            captured_piece = figure.capture(goal_field)
        except FieldMustBeCastledError:
            captured_piece = figure.castle(goal_field)
        moved_piece = {"figure": figure, "old_position": start_field}

        if self.king.in_check:
            self.rollback(moved_piece, captured_piece)
            raise IllegalMoveError("Move would put player's king in check.")

        figure.already_moved = True
        figure.last_moved = self.chessboard.current_move
        if captured_piece:
            captured_piece["figure"].last_moved = self.chessboard.current_move
        self.chessboard.end_turn()



# -*- coding: utf-8 -*-
"""A queen chess piece."""

from .figure import Figure


class Queen(Figure):

    """A queen chess piece."""

    start_position = {"white": ["D1"],
                      "black": ["D8"]}

    def __str__(self):
        if self.color == "white":
            string = "♕"
        else:
            string = "♛"
        return string

    @property
    def legal_moves(self):
        """Return legal moves from current position."""
        return self._fields_in_directions(("to_left", "to_right", "above",
                                           "below", "above_right",
                                           "below_right", "above_left",
                                           "below_left"))



"""
Chess figures.

Import all figure types into top module.
"""

from bishop import Bishop
from king import King
from knight import Knight
from pawn import Pawn
from queen import Queen
from rook import Rook



"""Abstract chess figure classes."""

from king_snake.errors import (FieldMustBeCastledError, FieldOccupiedError,
                               IllegalCaptureError, IllegalMoveError,
                               PawnMustCaptureError)


class Figure(object):

    """
    A chess figure.

    start_position is the starting position for the figure class and should be
    implemented in subclasses. It contains a dictionary of possible positions
    for each color. Each entry in the dictionary is a tuple of the appropriate
    field coordinates.
    """

    start_position = {"white": (None),
                      "black": (None)}

    def __repr__(self):
        return "{type} at {position}".format(type=self.__class__.__name__,
                                             position=self.position)

    def __init__(self, player, position=None):
        """
        Initialize figure and position figure on chessboard.

        If no initialization position is given, the start position is
        determined by the player's color and whether or not the position on the
        board is already taken.

        A Player set's a figure's already_moved field to True when moving. The
        last_moved field is set to the current move number whenever a move is
        successfully completed. A captured move is also set to have been moved
        when being captured, as is a castled rook.
        """
        self.player = player
        if self.player == self.player.chessboard.players["white"]:
            self.color = "white"
        else:
            self.color = "black"

        self.position = position
        if not position:
            self._set_start_position()

        self.already_moved = False
        self.last_moved = None

    def _set_start_position(self):
        """
        Set figure's starting position.

        Go through all possible positions for object's color. Try to take
        position. If position is blocked, try next position.
        """
        for position in self.start_position[self.color]:
            if not self.position:
                try:
                    self.player.chessboard.fields[
                                              position].receive_figure(self)
                    self.position = self.player.chessboard.fields[position]
                except FieldOccupiedError:
                    pass

    def _fields_in_directions(self, directions, perimeter=8):
        """
        Find fields within a perimeter in a given direction.

        Returns a list of fields in the given directions within the movement
        perimeter. Terminates after appending a field with a piece on it.

        The directions are the string names of directional methods on the Field
        class which return neighboring Fields in the given direction. See Field
        for the possible directions.

        The perimeter is set to 8 by default, because chess pieces operate on
        an 8x8 board.
        """
        fields = []
        for direction in directions:
            found_figure = None
            next_position = getattr(self.position, direction)()
            distance = 1
            while distance <= perimeter and next_position and not found_figure:
                fields.append(next_position)
                if next_position.figure:
                    found_figure = next_position.figure
                next_position = getattr(next_position, direction)()
                distance += 1
        return fields

    @property
    def legal_moves(self):
        """
        Compute list of legal moves.

        If legal, append field. If piece, append field and stop.
        """
        raise NotImplementedError

    def move(self, field):
        """
        Move to field.

        If field in self.legal_moves, add self to field, then replace
        self.position with new field. Else raise FieldOccupiedError.
        """
        if field in self.legal_moves:
            field.receive_figure(self)
        else:
            raise IllegalMoveError("Piece cannot move to given field.")

    def capture(self, field):
        """Capture piece at field.

        Check if piece to capture is of different color. If not, raise error.
        If it is, unlink the figure and field, then place self on field.

        @return dict: {"figure": figure, "old_position": old_position}
        """
        if field.figure.color != self.color:
            captured_figure = field.figure
            captured_figure.position = None
            field.figure = None
            field.receive_figure(self)
            return {"figure": captured_figure, "old_position": field}
        else:
            raise IllegalCaptureError("Cannot capture piece of same color.")



# -*- coding: utf-8 -*-
"""Pawn chess piece."""

from .figure import Figure, IllegalCaptureError, PawnMustCaptureError
from .queen import Queen


class Pawn(Figure):

    """Pawn chess piece."""

    start_position = {"white": list(letter + "2" for letter in "ABCDEFGH"),
                      "black": list(letter + "7" for letter in "ABCDEFGH")}

    def __str__(self):
        if self.color == "white":
            string = "♙"
        else:
            string = "♟"
        return string

    def __init__(self, player):
        """Initialize Pawn, set direction and change thresholds by color."""
        super(Pawn, self).__init__(player)
        self.en_passant_ready = False
        self.can_be_taken_en_passant = True
        if self.color == "white":
            self.move_direction = "above"
            self.last_row = 8
            self.en_passant_row = 5
            self.first_jump_row = 4
        else:
            self.move_direction = "below"
            self.last_row = 1
            self.en_passant_row = 4
            self.first_jump_row = 5

    def _finish_turn(self):
        """
        Monitor pawn state changes.

        If Pawn reaches last row, it is exchanged for a Queen. If Pawn is in
        its en passant row, it is ready to perform en passant. If Pawn moves
        two squares on its first move, it can be taken by en passant.
        """

        if self.position.number == self.last_row:
            # Remove self from field and place queen there
            self.position.figure = None
            queen = Queen(self.player, self.position)
            self.position.figure = queen
            # Replace self in player's list with queen
            self.player.figures[self.player.figures.index(self)] = queen

        elif self.position.number == self.en_passant_row:
            self.en_passant_ready = True

        elif (self.position.number == self.first_jump_row and not
              self.already_moved):
            self.can_be_taken_en_passant = True

    @property
    def legal_moves(self):
        """
        Return legal moves from current position.

        If pawn has already moved, it has a range of one square. Otherwise it
        has a range of two squares. Also, if the adjacent squares in the pawn's
        capture directions contain players, those fields are added to the legal
        moves.
        """
        if self.color == "white":
            movement_direction = "above"
            capture_directions = "above_left", "above_right"
        else:
            movement_direction = "below"
            capture_directions = "below_left", "below_right"

        if self.already_moved:
            move_range = 1
        else:
            move_range = 2
        moves = self._fields_in_directions([movement_direction], move_range)
        for field in self._fields_in_directions(capture_directions, 1):
            if field and self.en_passant_ready or field.figure:
                moves.append(field)
        return moves

    def move(self, field):
        """
        Move to given field.

        can_be_taken_en_passant is always set to False in order to override
        previous settings to true. Otherwise that state would be carried over
        even when it was no longer true.
        """

        self.can_be_taken_en_passant = False
        if (field in self.legal_moves and
                field in self._fields_in_directions([self.move_direction], 2)):
            super(Pawn, self).move(field)
            self._finish_turn()
        else:
            raise PawnMustCaptureError("Pawn can move diagonally only when "
                                       "capturing.")

    def capture(self, field):
        """Capture piece if legal, diagonally adjacent or en passant."""
        if self.color == "white":
            capture_directions = ("above_left", "above_right")
            en_passant_direction = "below"
        else:
            capture_directions = ("below_left", "below_right")
            en_passant_direction = "above"

        if field not in self._fields_in_directions(capture_directions, 1):
            raise IllegalCaptureError("Pawns can only capture forward "
                                      "diagonally.")

        # En passant
        if not field.figure and self.en_passant_ready:
            field.receive_figure(self)
            capture_position = self._fields_in_directions(
                                                 [en_passant_direction], 1)[0]
            capture_figure = capture_position.figure
            if (isinstance(capture_figure, Pawn) and
                capture_figure.can_be_taken_en_passant and
                capture_figure.last_moved == (
                                  self.position.chessboard.current_move - 1)):
                capture_dict = super(Pawn, self).capture(capture_position)
                field.receive_figure(self)
        # Normal capture
        else:
            capture_dict = super(Pawn, self).capture(field)
        self._finish_turn()
        return capture_dict
